
from enum import Enum

class HostProtocol(str, Enum):
    WEBSOCKET = "WEBSOCKET"

class ProviderProtocol(str, Enum):
    WEBSOCKET = "WEBSOCKET"

class PostmanProtocol(str, Enum):
    WEBSOCKET = "WEBSOCKET"
    KAFKA = "KAFKA"
    RABBITMQ = "RABBITMQ"

class ClientType(str, Enum):
    HOST = "HOST"
    CLIENT = "CLIENT"
    PROVIDER = "PROVIDER"
    POINT = "POINT"


class DataPointType(str, Enum):
    GRAPHQL = "GRAPHQL"
    REST = "REST"



class TYPENAMES(str, Enum):
    MODELPORTTYPE = "ModelPortType"


from typing import Any, Callable, Mapping
import logging

logger = logging.getLogger(__name__)

class ArgPortExtender:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._widget = None

    def buildWidget(self, **kwargs):
        try:
            from bergen.ui.widgets import QueryWidget, SliderWidget, IntWidget, VoidWidget, BaseWidgetMixin

            typeWidgetMap: Mapping[str, Callable[[Any], BaseWidgetMixin]] = {
                "QueryWidgetType": lambda port: QueryWidget.fromPort(port, **kwargs),
                "SliderWidgetType": lambda port: SliderWidget.fromPort(port, **kwargs),
                "IntWidgetType": lambda port: IntWidget.fromPort(port, **kwargs),
            }

            widgetBuilder = typeWidgetMap.get(self.widget.TYPENAME, lambda port: VoidWidget.fromPort(port, **kwargs))
            self._widget = widgetBuilder(self)
            return self._widget

        except ImportError as e:
            logger.error("Please install PyQt5 in Order to use widgets")
            raise e
    
    def __call__(inputs: dict, provider="vard"):
        logger.info("Called with inputs")

        logger.info("Serializing Inputs")


    def _repr_html_(self):
        string = f"{self.TYPENAME} with {self.key}"

        return string


class KwargPortExtender:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._widget = None

    def buildWidget(self, **kwargs):
        try:
            from bergen.ui.widgets import QueryWidget, SliderWidget, IntWidget, VoidWidget, BaseWidgetMixin

            typeWidgetMap: Mapping[str, Callable[[Any], BaseWidgetMixin]] = {
                "QueryWidgetType": lambda port: QueryWidget.fromPort(port, **kwargs),
                "SliderWidgetType": lambda port: SliderWidget.fromPort(port, **kwargs),
                "IntWidgetType": lambda port: IntWidget.fromPort(port, **kwargs),
            }

            widgetBuilder = typeWidgetMap.get(self.widget.TYPENAME, lambda port: VoidWidget.fromPort(port, **kwargs))
            self._widget = widgetBuilder(self)
            return self._widget

        except ImportError as e:
            logger.error("Please install PyQt5 in Order to use widgets")
            raise e
    
    def __call__(inputs: dict, provider="vard"):
        logger.info("Called with inputs")

        logger.info("Serializing Inputs")


    def _repr_html_(self):
        string = f"{self.TYPENAME} with {self.key}"

        return string


class ReturnPortExtender:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


    def _repr_html_(self):
        string = f"{self.TYPENAME} with {self.key}"

        return string


from typing import Dict
from bergen.queries.initiation import HOST_GQL
from bergen.schema import DataModel
from bergen.enums import ClientType, DataPointType
from bergen.clients.default import Bergen


class PointBergen(Bergen):

    def __init__(self, *args, point_inward=None, point_outward=None, point_port=8000, point_type = DataPointType.GRAPHQL, scopes= [],**kwargs) -> None:
        super().__init__(*args, client_type = ClientType.POINT, scopes=scopes + ["host","provide"], **kwargs)
        self.model_registry: Dict[str, DataModel] = {}
        self.negotiate_additionals = {
            "inward": point_inward,
            "outward": point_outward,
            "port": point_port,
            "pointType": point_type.value
        }


    def register(self, model: DataModel):
        self.model_registry[model.identifier] = model


    def start(self):
        if self.loop.is_running() and not self.running_in_sync: 
             raise NotImplementedError("We cannot start in an Eventloop")
        else:
            self.negotiate()

        params = []
        for model in self.model_registry.values():
            params = {"identifier": model.identifier, "extenders": model.extenders}
            HOST_GQL.run(ward=self.main_ward, variables=params)

        

     




from typing import List, Optional
from pydantic import BaseModel


class ProvideParams(BaseModel):
    providers: Optional[List[str]]
    auto_unprovide: Optional[bool] = True
    auto_provide: Optional[bool] = True



from ...messages.base import MessageDataModel, MessageMetaModel,MessageModel

from typing import Optional
from ...messages.types import DEACTIVATE_POD


class MetaModel(MessageMetaModel):
    type: str = DEACTIVATE_POD

class DataModel(MessageDataModel):
    pod: Optional[int]


class DeActivatePodMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from pydantic.main import BaseModel
from ....messages.types import  RESERVE, UNRESERVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional



class ProvideMetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class ProvideMetaModel(MessageMetaModel):
    type: str = UNRESERVE
    extensions: Optional[ProvideMetaExtensionsModel]

class DataModel(MessageDataModel):
    reservation: str



class UnreserveMessage(MessageModel):
    data: DataModel
    meta: ProvideMetaModel


from ....messages.generics import Token
from pydantic.main import BaseModel
from ....messages.types import  BOUNCED_ASSIGN
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional



class AssignParams(BaseModel):
    providers: Optional[List[str]]

class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = BOUNCED_ASSIGN
    extensions: Optional[MetaExtensionsModel]
    token: Token

class DataModel(MessageDataModel):
    reservation: str # The reservation reference we are going to assign to
    args: dict
    kwargs: Optional[dict]

class BouncedAssignMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from pydantic.main import BaseModel
from .params import ProvideParams
from ....messages.types import  PROVIDE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional

class ProvideMetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class ProvideMetaModel(MessageMetaModel):
    type: str = PROVIDE
    extensions: Optional[ProvideMetaExtensionsModel]

class ProvideDataModel(MessageDataModel):
    node: Optional[str] #TODO: Maybe not optional
    template: Optional[str]
    params: Optional[ProvideParams]


class ProvideMessage(MessageModel):
    data: ProvideDataModel
    meta: ProvideMetaModel


from asyncio.tasks import ensure_future
from bergen.messages import *
from bergen.messages.utils import expandToMessage 
from typing import Callable
from bergen.utils import expandOutputs, shrinkInputs
from bergen.messages.exception import ExceptionMessage
from bergen.messages.base import MessageModel
from bergen.postmans.base import BasePostman
import uuid
import logging
from functools import partial
import json
from bergen.console import console
import asyncio
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


import websockets
from bergen.schema import AssignationStatus, Node, ProvisionStatus, Template
from bergen.models import Pod



logger = logging.getLogger(__name__)


class NodeException(Exception):
    pass

class PodException(Exception):
    pass

class ProviderException(Exception):
    pass


class Channel:

    def __init__(self, queue) -> None:
        self.queue = queue
        pass


class WebsocketPostman(BasePostman):
    type = "websocket"

    def __init__(self, port= None, protocol = None, host= None, auth= None, **kwargs) -> None:
        self.token = auth["token"]
        self.port = port
        self.protocol = protocol
        self.host = host
        self.connection = None      
        self.channel = None         
        self.callback_queue = ''

        self.uri = f"ws://{self.host}:{self.port}/postman/?token={self.token}"
        self.progresses = {}

        # Retry logic
        self.allowed_retries = 2
        self.current_retries = 0

        # Result and Stream Function
        self.futures = {}
        self.streams = {}   # Are queues that are consumed by tasks
        
        # Progress
        self.progresses = {}  # Also queues
        self.pending = []


        self.assign_routing = "assignation_request"
        super().__init__(**kwargs)

    async def connect(self):
        self.callback_queue = asyncio.Queue()
        self.progress_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()


        self.tasks = []

        self.startup_task = create_task(self.startup())


    async def disconnect(self):

        for task in self.pending:
            task.cancel()

        if self.connection: await self.connection.close()
        if self.receiving_task: self.receiving_task.cancel()
        if self.sending_task: self.sending_task.cancel()
        if self.callback_task: self.callback_task.cancel()

        if self.startup_task:
            self.startup_task.cancel()

        try:
            await self.startup_task
        except asyncio.CancelledError:
            logger.info("Postman disconnected")

    async def startup(self):
        try:
            await self.connect_websocket()
        except Exception as e:
            logger.debug(e)
            self.current_retries += 1
            if self.current_retries < self.allowed_retries:
                sleeping_time = (self.current_retries + 1)
                logger.error(f"Connection to Arkitekt Failed: Trying again in {sleeping_time} seconds.")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                return

        self.receiving_task = create_task(
            self.receiving()
        )

        self.sending_task = create_task(
            self.sending()
        )

        self.callback_task = create_task(
            self.callbacks()
        )


        done, self.pending = await asyncio.wait(
            [self.callback_task, self.receiving_task, self.sending_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        exceptions = [ task.exception() for task in done]
        try:
            for e in exceptions:
                raise e
        except:
            console.print_exception()


        logger.debug(f"Postman: Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Postman: Trying to reconnect Postman')
        console.print()

        if self.connection: await self.connection.close()

        for task in self.pending:
            task.cancel()

        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup()


    async def connect_websocket(self):
        self.connection = await websockets.client.connect(self.uri)
        logger.info("Successfully connected [bold]Postman")
        

    async def receiving(self):
        async for message in self.connection:
            await self.callback_queue.put(message)
    
    async def sending(self):
        while True:
            message = await self.send_queue.get()
            if self.connection:
                await self.connection.send(message.to_channels())
            else:
                raise Exception("No longer connected. Did you use an Async context manager?")

            self.send_queue.task_done()

    async def callbacks(self):
        while True:
            message = await self.callback_queue.get()
            try:
                parsed_message = expandToMessage(json.loads(message))
                await self.on_message(parsed_message)
            except:
                raise 

            self.callback_queue.task_done()


    async def forward(self, message: MessageModel):
        await self.send_queue.put(message)
        






from bergen.types.node.widgets.base import BaseWidget


class StringWidget(BaseWidget):

    def __init__(self, min=None, max=None, **kwargs) -> None:
        super().__init__(**kwargs)
        #TODO: Inspect if widgets dependencies are okay for that query



from .bounced_reserve import BouncedReserveMessage
from .reserve_critical import ReserveCriticalMessage
from .reserve_done import ReserveDoneMessage
from .reserve_progress import ReserveProgressMessage
from .reserve import ReserveMessage
from .bounced_forwarded_reserve import BouncedForwardedReserveMessage


from .host import DeActivatePodMessage, ActivatePodMessage
from .postman.assign import *
from .base import MessageModel
from .types import *
from .postman.provide import *
from .postman.unprovide import *
from .postman.reserve import *
from .postman.unreserve import *
from .postman.unassign import *
from .exception import ExceptionMessage
import json


registry = {
    PROVIDE_DONE:  ProvideDoneMessage,
    BOUNCED_PROVIDE: BouncedProvideMessage,
    PROVIDE_CRITICAL: ProvideCriticalMessage,
    PROVIDE_PROGRESS: ProvideProgressMessage,
    PROVIDE: ProvideMessage,

    UNPROVIDE_DONE:  UnprovideDoneMessage,
    BOUNCED_UNPROVIDE: BouncedUnprovideMessage,
    UNPROVIDE_CRITICAL: UnprovideCriticalMessage,
    UNPROVIDE_PROGRESS: UnprovideProgressMessage,
    UNPROVIDE: UnprovideMessage,

    RESERVE: ReserveMessage,
    BOUNCED_RESERVE: BouncedReserveMessage,
    BOUNCED_FORWARDED_RESERVE: BouncedForwardedReserveMessage,
    RESERVE_CRITICAL: ReserveCriticalMessage,
    RESERVE_PROGRESS: ReserveProgressMessage,
    RESERVE_DONE: ReserveDoneMessage,

    UNRESERVE: UnreserveMessage,
    BOUNCED_UNRESERVE: BouncedUnreserveMessage,
    UNRESERVE_CRITICAL: UnreserveCriticalMessage,
    UNRESERVE_PROGRESS: UnreserveProgressMessage,
    UNRESERVE_DONE: UnreserveDoneMessage,

    ASSIGN: AssignMessage,
    BOUNCED_ASSIGN: BouncedAssignMessage,
    BOUNCED_FORWARDED_ASSIGN: BouncedForwardedAssignMessage,
    ASSIGN_CRITICAL: AssignCriticalMessage,
    ASSIGN_PROGRES: AssignProgressMessage,
    ASSIGN_RETURN: AssignReturnMessage,
    ASSIGN_DONE: AssignDoneMessage,
    ASSIGN_YIELD: AssignYieldsMessage,

    UNASSIGN: UnassignMessage,
    BOUNCED_UNASSIGN: BouncedUnassignMessage,
    UNASSIGN_CRITICAL: UnassignCriticalMessage,
    UNASSIGN_PROGRES: UnassignProgressMessage,
    UNASSIGN_DONE: UnassignDoneMessage,

    ACTIVATE_POD: ActivatePodMessage,
    DEACTIVATE_POD: DeActivatePodMessage,

    EXCEPTION: ExceptionMessage
}


class MessageError(Exception):
    pass


def expandToMessage(message: dict) -> MessageModel:
    assert isinstance(message, dict), "Please provide already serialized Messages"
    try:
        cls: MessageModel = registry[message["meta"]["type"]]
    except:
        raise MessageError(f"Didn't find an expander for message {message} {message['meta']['type']}")

    return cls(**message)
    

def expandFromRabbitMessage(message) -> MessageModel:
    text = message.body.decode()
    return expandToMessage(json.loads(text))
